/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * actor.c: Manage actors and loading/saving games
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "actor.h"
#include "board.h"
#include "imath.h"

actor *cplayer = NULL;
actor *actors[MAX_ACTORS];
int num_actors = 0;
int num_actors_allocated = 0;
objectgroup actorclasses;
objectgroup hookfuncs;

void actor_load(void* ob, buffer * buf) {
    actor* o = (actor*)ob;
    
    o->vx = o->vy = 0;
    o->x = reads32(buf);
    o->y = reads32(buf);
    o->ang = readu16(buf);
}

void actor_serial(void* ob, buffer* buf, int phase) {
    actor* self = (actor*)ob;

    S_HOOK(onremove);
    S_TAGSET(tags);
    S_INT(x);
    S_INT(y);
    S_INT(vx);
    S_INT(vy);
    S_INT(w);
    S_INT(h);
    S_INT(ang);
}

void actor_init(void* self) {
    tagset_init(&((actor*)self)->tags);
    ((actor*)self)->onremove = NULL;
    ((actor*)self)->x = 0;
    ((actor*)self)->y = 0;
    ((actor*)self)->ang = 0;
    ((actor*)self)->bnext = -1;
}

void actor_fini(void* self) {
    tagset_clear(&((actor*)self)->tags);
    free_hook(((actor*)self)->onremove);
    ((actor*)self)->onremove = NULL;
}

int actor_add(void* obj) {
    if (num_actors == MAX_ACTORS)
	return 0;
    actors[num_actors++] = (actor*)obj;
    return 1;
}

void actor_remove(void* o) {
    actor* a = (actor*)o;
    call_hook(a->onremove);
    free_hook(a->onremove);
    a->onremove = NULL;
    CALLMETH((actor*)o, fini);
    ADDTAG((actor*)a, actor_remove);
}

void register_hook(hook** h, hookfunc* hf, const char* fmt, ...) {
    int nparams = RB strlen(fmt);
    int i;
    va_list lst;
    va_start(lst, fmt);
    
    hookparam* cp;
    hook* newhook = (hook*)malloc(sizeof(hook) + (sizeof(hookparam) * nparams));
    newhook->hf = hf;
    newhook->next = *h;
    *h = newhook;
    newhook->nparams = nparams;
    cp = &newhook->params[0];

    for (i = 0; i < nparams; i++, cp++, fmt++) {
        switch (*fmt) {
        case 'i':
            cp->type = PT_INT;
            cp->ival = va_arg(lst, int);
            break;
        case 's':
            cp->type = PT_STRING;
            cp->sval = strdup(va_arg(lst, char*));
            break;
        case 'a':
            cp->type = PT_ACTOR;
            cp->act = va_arg(lst, actor*);
            object_ref(cp->act);
            break;
        }
    }
    
    va_end(lst);
}

static void free_hook_params(hook* h) {
    int i;
    for (i = 0; i < h->nparams; i++) {
        hookparam* hp = &h->params[i];
        switch (hp->type) { 
        case PT_INT:
            break;
        case PT_STRING:
            free(hp->sval); 
            break;
        case PT_ACTOR:
            object_deref(hp->act);
            break;
        }
    }
}

void free_hook(hook* h) {
    hook* old;
    while (h) {
        old = h;
        h = h->next;
        free_hook_params(old);
        free(old);
    }
}

void call_hook(hook* h) {
    while (h) {
        if (h->hf)
            h->hf->func(h);
        h = h->next;
    }
}

void serial_actor_ref(void* a, buffer* buf, int phase) {
    actor* o;
    switch (phase) {
    case PHASE_WRITE:
        o = (actor*)*((void**)a);
        if (o) {
            if (o->bnext == -1) {
                o->bnext = num_actors;
                actors[num_actors++] = o;
                object_ref(o);
            }
            writeu16(buf, o->bnext+1);
        } else {
            writeu16(buf, 0);
        }
        break;
    case PHASE_READ:
        *((void**)a) = (void*)(uint32)readu16(buf);
        break;
    case PHASE_LINK:
        if (*((void**)a)) {
            *((void**)a) = actors[*((uint32*)a) - 1];
            object_ref(*((void**)a));
        }
        break;
    }
}

void write_hook_params(buffer* buf, hook* h) {
    if (!h->hf) {
        writeu16(buf, 0);
        writeu8(buf, 0);
        return;
    }

    writestr(buf, h->hf->name);
    writeu8(buf, h->nparams);
    int i;
    for (i = 0; i < h->nparams; i++) {
        hookparam* hp = &h->params[i];
        writeu8(buf, hp->type);
        switch (hp->type) {
        case PT_INT:
            writeu32(buf, hp->ival);
            break;
        case PT_STRING:
            writestr(buf, hp->sval);
            break;
        case PT_ACTOR:
            serial_actor_ref(&hp->act, buf,  0);
            break;
        }
    }
}

hook* read_hook_params(buffer* buf) {
    hookfunc* hf = (hookfunc*)read_namedobject(&hookfuncs, buf);
    int nparams = readu8(buf);
    int i;
    hookparam* cp;
    hook* newhook = (hook*)malloc(sizeof(hook) + (sizeof(hookparam) * nparams));
    newhook->hf = hf;
    newhook->next = NULL;
    newhook->nparams = nparams;
    cp = &newhook->params[0];
    for (i = 0; i < nparams; i++, cp++) {
        cp->type = readu8(buf);
        switch (cp->type) {
        case PT_INT:
            cp->ival = readu32(buf);
            break;
        case PT_STRING:
            readstr_alloc(buf, &cp->sval);
            break;
        case PT_ACTOR:
            serial_actor_ref(&cp->act, buf, 1);
            break;
        }
    }
    return newhook;
}

void serial_hook(hook** hp, buffer* buf, int phase) {
    hook* h;
    hookparam* cp;
    int c, i;

    switch (phase) {
    case PHASE_WRITE:
        h = *hp;
        c = 0;
        while (h) {
            c++; 
            h = h->next;
        }
        writeu8(buf, c);
        h = *hp;
        while (h) {
            write_hook_params(buf, h);
            h = h->next;
        }
        break;
    case PHASE_READ:
        *hp = NULL;
        c = readu8(buf);
        for (i = 0; i < c; i++) {
            h = read_hook_params(buf);
            *hp = h;
            hp = &h->next;
        }

        break;
    case PHASE_LINK:
        h = *hp;
        while (h) {
            cp = &h->params[0];
            for (i = 0; i < h->nparams; i++, cp++) {
                if (cp->type == PT_ACTOR) {
                    serial_actor_ref(&cp->act, buf, 2);
                }
            }
            h = h->next;
        }
        break;
    }
}



void clear_actors(void) {
    int i;
    for (i = 0; i < num_actors; i++) {
        actor* ca = actors[i];
        CALLMETH(ca, fini);
        object_deref(ca);
    }
    num_actors = 0;
}

void write_actors(buffer* buf) {
    int orig_num_actors = num_actors;
    int total_actors_pos;
    int i;
    for (i = 0; i < num_actors; i++) {
        actor* ca = actors[i];
        ca->bnext = i;
    }
    writeu16(buf, num_actors);
    total_actors_pos = buf->pos;
    writeu16(buf, 0);
    for (i = 0; i < num_actors; i++) {
        actor* ca = actors[i];
        int cpos = buf->pos;
        writeu16(buf, 0);
        writestr(buf, ca->class->name);

        actor_serial(ca, buf, 0);
        CALLMETH(ca, serial, buf, 0);
        int npos = buf->pos;
        buf->pos = cpos;
        writeu16(buf, (npos - cpos) - 2);
        buf->pos = npos;
    }
    
    i = buf->pos;
    buf->pos = total_actors_pos;
    writeu16(buf, num_actors);
    buf->pos = i;

    for (i = 0; i < num_actors; i++) {
        actor* ca = actors[i];
        ca->bnext = -1;
    }
    for (i = orig_num_actors; i < num_actors; i++) {
        actor* ca = actors[i];
        object_deref(ca);
    }
    num_actors = orig_num_actors;
}

void read_actors(buffer* buf) {
    clear_actors();
    num_actors = readu16(buf);
    int real_num_actors = readu16(buf);
    int i;
    for (i = 0; i < real_num_actors; i++) {
        actorclass* class;
        actor* o;
        int nextpos = readu16(buf);
        nextpos += buf->pos;
        class = (actorclass*)read_namedobject(&actorclasses, buf);
        if (class) {
            o = (actor*)_object_new((baseclass*)class);
            actors[i] = o;
            
            actor_serial(o, buf, 1);
            CALLMETH(o, serial, buf, 1);
            
            if (HASTAG(o, is_player)) {
                object_deref(cplayer);
                cplayer = o;
                object_ref(cplayer);
            }
        } else {
            actors[i] = NULL;
        }
        buf->pos = nextpos;
    }

    for (i = 0; i < real_num_actors; i++) {
        actor* o = actors[i];
        if (o) {
            actor_serial(o, buf, 2);
            CALLMETH(o, serial, buf, 2);
        }
    }

    for (i = num_actors; i < real_num_actors; i++) {
        actor* ca = actors[i];
        object_deref(ca);
    }
    for (i = 0; i < num_actors; i++) {
        if (!actors[i]) {
            actors[i--] = actors[--num_actors];
        }
    }
}

int save_game(const char* fname, const char* title) {
    buffer buf;
    buf.cap = 24576 + 32;
    buf.len = 0;
    buf.pos = 0;
    buf.data = malloc(buf.cap);

    RB strncpy(buf.data, title, 31);
    buf.data[31] = 0;

    buf.pos = 32;

    save_board_to(&buf);
    write_actors(&buf);
    
    int ret = writefile(fname, &buf);

    free(buf.data);
    
    return ret;
}

int get_savegame_title(const char* fname, char* title) {
    int fd;

    fd = RB open(fname, O_RDONLY);

    if (fd < 0) {
        return fd;
    }

    RB read(fd, title, 31);
    title[31] = 0;

    RB close(fd);

    return 0;
}

int load_game(const char* fname) {
    buffer buf;
    buf.data = NULL;
    buf.pos = 0;
    buf.len = 0;
    buf.cap = 0;
    int ret = loadfile(fname, &buf);

    if (ret < 0) {
        if (buf.data) free(buf.data);
        return ret;
    }

    buf.pos = 32;

    load_board_from(&buf);
    read_actors(&buf);
    
    free(buf.data);
    return 0;
}

inline int inty(int x1, int y1, int x2, int y2, int hx)  ICODE_ATTR;
inline int inty(int x1, int y1, int x2, int y2, int hx) {
    int vx, vy;
    vx = x2 - x1;
    vy = y2 - y1;
    return y1 + div32((vy * (hx - x1)), vx);
}
inline int intx(int x1, int y1, int x2, int y2, int hy)  ICODE_ATTR;
inline int intx(int x1, int y1, int x2, int y2, int hy) {
    int vx, vy;
    vx = x2 - x1;
    vy = y2 - y1;
    return x1 + div32((vx * (hy - y1)), vy);
}

int clipr(int x1, int y1, int w1, int h1, int ox2, int oy2, int *x2, int *y2,
	  int w2, int h2)  ICODE_ATTR;
int clipr(int x1, int y1, int w1, int h1, int ox2, int oy2, int *x2,
	  int *y2, int w2, int h2) {
    int tox = ox2 - x1;
    int toy = oy2 - y1;
    int tnx = *x2 - x1;
    int tny = *y2 - y1;
    int tw = w1 + w2 - 1;
    int th = h1 + h2 - 1;

    if (tox < -tw && tnx < -tw)
	return 0;
    if (tox > tw && tnx > tw)
	return 0;
    if (toy < -th && tny < -th)
	return 0;
    if (toy > th && tny > th)
	return 0;

    if (tox <= -tw && tnx > -tw) {
	int hy = inty(tox, toy, tnx, tny, -tw);
	if (hy >= -th && hy <= th) {
	    *x2 = x1 - tw - 1;
	    return 1;
	}
    }

    if (tox >= tw && tnx < tw) {
	int hy = inty(tox, toy, tnx, tny, tw);
	if (hy >= -th && hy <= th) {
	    *x2 = x1 + tw + 1;
	    return 1;
	}
    }

    if (toy <= -th && tny > -th) {
	int hx = intx(tox, toy, tnx, tny, -th);
	if (hx >= -tw && hx <= tw) {
	    *y2 = y1 - th - 1;
	    return 2;
	}
    }

    if (toy >= th && tny < th) {
	int hx = intx(tox, toy, tnx, tny, th);
	if (hx >= -tw && hx <= tw) {
	    *y2 = y1 + th + 1;
	    return 2;
	}
    }
    return 0;
}

int clipb(int bx, int by, int ox, int oy, int *x, int *y, int w, int h) {
    if (BCELL(bx, by) & 32768)
	return clipr((bx << 16) | 32768, (by << 16) | 32768, 32768, 32768, ox,
		     oy, x, y, w, h);
    return 0;
}

int _actor_move(actor* o, int vx, int vy) {
    int ret = 0;
    int ox = o->x;
    int oy = o->y;
    int bx, by, tx, ty, nx, ny, j;
    nx = o->x + vx;
    ny = o->y + vy;

    bx = o->x >> 16;
    by = o->y >> 16;

    ret |= clipb(bx, by - 1, o->x, o->y, &nx, &ny, o->w, o->h);
    ret |= clipb(bx, by + 1, o->x, o->y, &nx, &ny, o->w, o->h);
    ret |= clipb(bx - 1, by, o->x, o->y, &nx, &ny, o->w, o->h);
    ret |= clipb(bx + 1, by, o->x, o->y, &nx, &ny, o->w, o->h);
    if (1) {
	ret |= clipb(bx - 1, by - 1, o->x, o->y, &nx, &ny, o->w, o->h);
	ret |= clipb(bx + 1, by - 1, o->x, o->y, &nx, &ny, o->w, o->h);
	ret |= clipb(bx + 1, by + 1, o->x, o->y, &nx, &ny, o->w, o->h);
	ret |= clipb(bx - 1, by + 1, o->x, o->y, &nx, &ny, o->w, o->h);
    }

    o->x = nx;
    o->y = ny;

    for (tx = -1; tx < 2; tx++) {
	for (ty = -1; ty < 2; ty++) {
	    for (j = BOBJ(bx + tx, by + ty); j != -1; j = actors[j]->bnext) {
		actor *o2 = actors[j];
		if (o == o2)
		    continue;
                if (CALLMETH(o, pushes, o2) && !HASTAG(o2, pushing)) {
                    ADDTAG(o, pushing);
                    int tr;
                    int tox = o2->x;
                    int toy = o2->y;
                    
                    _actor_move(o2, vx, vy);
                    tr = _actor_move(o2, tox - o2->x, toy - o2->y);
                    if (tr)
                        CALLMETH(o2, pushedby, o, tr);
                    DELTAG(o, pushing);
                }
                if (CALLMETH(o2, clips, o)) {
		    ret |=
			clipr(o2->x, o2->y, o2->w, o2->h, ox, oy, &nx,
			      &ny, o->w, o->h);
		}
	    }
	}
    }
    o->x = nx;
    o->y = ny;


    return ret;
}

int actor_move(void* obj) {
    actor* o = (actor*)obj;
    return _actor_move(o, o->vx, o->vy);
}

void move_actors(void) {
    int i, j, bx, by, ox, oy;

    for (i = 0; i < num_actors; i++) {
        actor* o = actors[i];
	if (!o || HASTAG(o, actor_remove)) {
            object_deref(o);
	    actors[i--] = actors[--num_actors];
	}
    }

    RB memset(boardactor, -1, 4096 * sizeof(short));

    for (i = 0; i < num_actors; i++) {
	actor *o = actors[i];
	short *bo = &BOBJ(o->x >> 16, o->y >> 16);
	o->bnext = *bo;
	*bo = i;
    }

    for (i = 0; i < num_actors; i++) {
	actor *o = actors[i];
	bx = o->x >> 16;
	by = o->y >> 16;
	for (ox = -1; ox <= 1; ox++)
	    for (oy = -1; oy <= 1; oy++) {
		for (j = BOBJ(bx + ox, by + oy); j != -1;
		     j = actors[j]->bnext) {
		    actor *o2 = actors[j];
		    if (o == o2)
			continue;

		    if (abs(o->x - o2->x) <= o->w + o2->w
			&& abs(o->y - o2->y) <= o->h + o2->h)
			CALLMETH(o,touch,o2);
		}
	    }

    }
    for (i = 0; i < num_actors; i++) {
	actor *o = actors[i];
	CALLMETH(o,move);
    }
}

void draw_actors(void) {
    int i;
    for (i = 0; i < num_actors; i++) {
	CALLMETH(actors[i],draw);
    }

}

actor* actor_create(buffer* b) {
    actorclass* class;
    if (b->pos >= b->len - 1) return NULL;

    class = (actorclass*)read_namedobject(&actorclasses, b);

    if (class) {
	actor *o;
	o = (actor*)_object_new((baseclass*)class);
	
	if (o) {
	    CALLMETH(o, load, b);
	}
	return o;
    }
    return NULL;
}

