/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef __ACTOR_H_
#define __ACTOR_H_

#include "common.h"
#include "class.h"
#include "util.h"
#include "tag.h"

/* The maximum number of actors on the board*/
#define MAX_ACTORS 512

struct actor;
struct hook;

typedef enum { DMG_NO, DMG_YES, DMG_ABSORB } dmg_result;
enum { PT_INT, PT_STRING, PT_ACTOR };

typedef struct {
    int type;
    union {
        int ival;
        char* sval;
        struct actor* act;
    };
} hookparam;

typedef struct {
    char* name;
    void (*func)(struct hook* h);
} hookfunc;

typedef struct hook {
    hookfunc* hf;
    struct hook* next;
    int nparams;
    hookparam params[0];
} hook;

typedef struct actorclass {
    CLASS_HEAD;
    void (*load) (void* self, buffer* buf);
    void (*move) (void* self);
    void (*draw) (void* self);
    int (*clips) (void* self, struct actor* other);
    int (*pushes) (void* self, struct actor* other);
    void (*pushedby) (void* self, struct actor* other, int res);
    void (*touch) (void* self, struct actor* other);
    int (*damage) (void* self, tagset* type, int amt);
    void (*serial) (void* self, buffer* buf, int type);
} actorclass;

#define ACTOR_HEAD                              \
    hook* onremove;                             \
    tagset tags;                                \
    int x, y;                                   \
    int vx, vy;                                 \
    int w;                                      \
    int h;                                      \
    int ang;                                    \
    int bnext

typedef struct actor {
    OBJECT_HEAD(actorclass);
    ACTOR_HEAD;
} actor;

extern actor *cplayer;
extern actor *actors[];
extern int num_actors;

extern int num_actors_allocated;

extern objectgroup actorclasses;
extern objectgroup hookfuncs;

actor* actor_create(buffer* b);

void actor_init(void* obj);
void actor_fini(void* obj);
void actor_load(void* o, buffer* b);
void actor_serial(void* ob, buffer* buf, int phase);
int  actor_add(void* obj);
void actor_remove(void* o);
int  actor_move(void* o);

int save_game(const char* fname, const char* title);
int get_savegame_title(const char* fname, char* title);
int load_game(const char* fname);
void clear_actors(void);

void free_hook(hook* h);
void call_hook(hook* h);
void register_hook(hook** hook, hookfunc* hf, const char* fmt, ...);

void serial_hook(hook** hp, buffer* buf, int phase);
void serial_actor_ref(void* a, buffer* buf, int phase);

#define SERIALIZER(clas) METHOD(void, clas, serial, buffer* buf, int phase)

#define S_BYTE(f) serial_u8(&self->f, buf, phase)
#define S_SHORT(f) serial_u16(&self->f, buf, phase)
#define S_INT(f) serial_u32(&self->f, buf, phase)
#define S_STR(f) serial_str(&self->f, buf, phase)
#define S_HOOK(f) serial_hook(&self->f, buf, phase)
#define S_ACTOR(f) serial_actor_ref(&self->f, buf, phase)
#define S_TAGSET(f) serial_tagset(&self->f, buf, phase)
#define S_BUFFER(f) serial_buffer(&self->f, buf, phase)


void move_actors(void);
void draw_actors(void);


#define ACTOR_CLASSDEF(nm, ...) _CLASSDEF(actorclass, nm, ACTOR_HEAD; __VA_ARGS__ );

#define ACTOR_INITCLASS(nm)                                     \
_INITCLASS(actorclass, nm, actor_init , actor_fini,             \
           (void (*)(void*, buffer *))actor_load,               \
           (void (*)(void*))_default_method,                    \
           (void (*)(void*))_default_method,                    \
           (int (*)(void*, actor *))_default_method,            \
           (int (*)(void*, actor *))_default_method,            \
           (void (*)(void*, actor *, int))_default_method,      \
           (void (*)(void*, actor *))_default_method,           \
           (int (*)(void*, tagset*, int))_default_method,       \
           (void (*)(void*, buffer*, int))_default_method);     \
REGISTER_NAMEDOBJECT(actorclasses,nm##_class);

#define ACTOR_ICLASSDEF(nm, ...)                \
ACTOR_CLASSDEF(nm, ## __VA_ARGS__);             \
ACTOR_INITCLASS(nm)



#define HOOK(name)                              \
static void __hook_##name(hook* h);             \
static hookfunc hook_##name = {                 \
    #name, __hook_##name                        \
};                                              \
REGISTER_NAMEDOBJECT(hookfuncs, hook_##name);   \
static void __hook_##name(hook* h)

#define HASTAG(obj, tag) tagset_in(&(obj)->tags, _TAG(#tag))
#define ADDTAG(obj, tag) tagset_add(&(obj)->tags, _TAG(#tag))
#define DELTAG(obj, tag) tagset_del(&(obj)->tags, _TAG(#tag))


#endif
