/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * malloc.c: Malloc implementation for RocKen
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/


/*

Memory allocation is done by splitting memory into a linked list of
chunks. Everything is aligned to an 8-byte boundary for simplicity and
speed. 

Memory chunks are tagged with an 8-byte header that gives the distance
to the next and previous boundary (in units of 8 bytes) and also if
the chunk is in use.

In addition, free chunks are added to a linked list sorted by
size. This ensures that any memory allocated is placed in the smallest
free chunk that will fit.

*/

/* define MALLOC_TEST to build the standalone tester */

#ifdef MALLOC_TEST
#define RB
#define ICODE_ATTR
#include <stdlib.h>
#else
#include "common.h"
#endif

typedef struct mallochead {
    /* The distance to the previous chunk, or 0 if at the beginning */
    unsigned int psize;
    /* The distance to the next chunk, or 0 if at the end. The high
       bit is used to mark the chunk allocated. */
    unsigned int size; 

    /* A linked list of free chunks*/
    struct mallochead *prev, *next;
} mallochead;

/* The start and size of the heap. Only needed for heap_dump. */
static mallochead *mallocstart;
static int heap_size;

/* The head node of the free chunk list. */
static mallochead freelist;

#define USE_BIT 0x80000000
#define SIZE_MASK 0x7fffffff
#define USED(m) ((m)->size&USE_BIT)
#define SIZE(m) ((m)->size&SIZE_MASK)
#define ADD(m, size) ((mallochead*)(((unsigned char*)(m)) + ((size)<<3)))
#define DIFF(a, b) ((((unsigned char*)(a)) - ((unsigned char*)(b)))>>3)

#define NEXT(m) ADD(m, (m)->size)
#define PREV(m) ADD(m, -(m)->psize)

/* Split a chunk at a specific size. The new chunks will be marked
   free, but will not be added to the free list. */
static inline mallochead* split_chunk(mallochead* start, int size) {
    mallochead* new = ADD(start, size);
    NEXT(start)->psize = new->size = start->size - size;
    start->size = new->psize = size;
    return new;
}

/* Remove a boundary between two free chunks. The chunk should be
   removed from the free list before joining. The combined chunk will
   not be added to the free list. */

static inline void join_chunks(mallochead* bound) {
    mallochead* start = PREV(bound);
    mallochead* end = NEXT(bound);
    end->psize = bound->psize + SIZE(bound);
    start->size = USED(start) | end->psize;
}

/* Add a free chunk to the free list. The chunk should be marked free
   before inserting. */

static inline void add_free_chunk(mallochead* chk) {
    unsigned int csize;
    mallochead* i;
    csize = chk->size;
    i = freelist.next;
    while (csize > i->size) i = i->next;
    chk->next = i;
    chk->prev = i->prev;
    i->prev->next = chk;
    i->prev = chk;
}

/* Remove a chunk from the free list. */

static inline void remove_free_chunk(mallochead* chk) {
    chk->prev->next = chk->next;
    chk->next->prev = chk->prev;
}

/* Finds the smallest free chunk that will fit the given size (in
   8-byte chunks). */
void malloc_heapdump(const char* fname);

static mallochead *find_free_chunk(unsigned int size) ICODE_ATTR;
static mallochead *find_free_chunk(unsigned int size) {
    mallochead* i;
    for (i = freelist.next; i != &freelist; i = i->next) {
        if (size <= i->size) return i;
    }
    return NULL;
}

/* Allocate a chunk of memory of at least _size_ bytes. Returns NULL
   if the memory could not be allocated. */

void *__malloc(unsigned int size) ICODE_ATTR;
void *__malloc(unsigned int size) {
    mallochead *m;

    /* round up to the nearest multiple of 8 and add 1 for the header */
    size = (size + 7) / 8 + 1;

    /* find a suitable free chunk */
    m = find_free_chunk(size);
    if (!m) {
#if 0
        static int x;
        if (!x) {
            debugf("Out of memory!!!!!!");
            malloc_heapdump("/foo.dump");
            x = 1;
        }
#endif
	return NULL;
    }
    remove_free_chunk(m);
    
    /* do we have any leftover space at the end? */
    if (size < (m->size - 1)) {
        mallochead* bound = split_chunk(m, size);
        add_free_chunk(bound);
    }

    /* mark it used */
    m->size |= USE_BIT;

    return (void *)ADD(m, 1);
}

/* Free a chunk of memory allocated by malloc. */

void __free(void *data) ICODE_ATTR;
void __free(void *data) {
    mallochead *m = ADD(data, -1);
    mallochead *n = NEXT(m);
    mallochead *p = PREV(m);

    m->size &= SIZE_MASK;

    if (n->size && !USED(n)) {
        remove_free_chunk(n);
        join_chunks(n);
    }

    if (m->psize && !USED(p)) {
        /* remove it and then re-insert it with its new size. */
        remove_free_chunk(p);
        join_chunks(m);
        m = p;
    }

    add_free_chunk(m);
}

/* Resize a chunk of allocated memory to at least _size_ bytes. Returns NULL
   if the memory could not be allocated. Attempts to keep the memory
   in the same place. */

void* __realloc(void *data, unsigned int size) ICODE_ATTR;
void* __realloc(void *data, unsigned int size) {
    mallochead *m = ADD(data, -1);
    mallochead *n = NULL, *p = NULL, *new = NULL;
    int mark_orig_free = 1;
    unsigned int datsize = size; /* number of bytes to copy */
    unsigned int csize, origsize;
    
    /* round up to the nearest multiple of 8 and add 1 for the header */
    size = (size + 7) / 8 + 1;

    origsize = csize = SIZE(m);

    if ((origsize - 1) * 8 < datsize) datsize = (origsize - 1) * 8;

    /* do we need to resize? */
    if (size == csize || size == (csize-1)) {
        return data;
    }
    
    n = NEXT(m);

    /* temporarily mark it free */
    m->size &= SIZE_MASK;

    /* if there is a free chunk after this one, join them */
    if (n->size && !USED(n)) {
        remove_free_chunk(n);
        join_chunks(n);
        
        /* check again with new size */
        csize = SIZE(m);
        if (size == csize || size == (csize-1)) {
            m->size |= USE_BIT;
            return data;
        }
    }

    /* shrink in place */
    if (size < csize) {
        n = split_chunk(m, size);
        add_free_chunk(n);

        m->size |= USE_BIT;

        return data;
    }

    /* if there is a free chunk *before* this one, join them */
    if (m->psize && !USED(p = PREV(m))) {
        mark_orig_free = 0;
        join_chunks(m);
        /* remove it and then re-insert it with its new size. */
        remove_free_chunk(p);
        add_free_chunk(p);
    }

    new = find_free_chunk(size);
    if (!new) {
        /* restore original chunks */
        if (m->psize && !USED(p)) {
            remove_free_chunk(p);
            split_chunk(p, DIFF(m, p));
            add_free_chunk(p);
        }
        if (origsize < (csize-1)) {
            new = split_chunk(m, origsize);
            add_free_chunk(new);
        }

        m->size |= USE_BIT;
        return NULL;
    }

    remove_free_chunk(new);

    if (new != m) /* sanity check */
	RB memmove(ADD(new, 1), data, datsize);

    /* if we didn't join with the previous chunk, add the original
       chunk to the free list. */
    if (mark_orig_free) {
        add_free_chunk(m);
    }

    /* leftover space at the end? */
    if (size < (new->size - 1)) {
        mallochead* bound = split_chunk(new, size);
        add_free_chunk(bound);
    }

    new->size |= USE_BIT;
    return ADD(new, 1);
}

/* Allocates memory for an array of _nmemb_ elements of _size_ bytes
   each and returns a pointer to the allocated memory. Returns NULL if
   memory could not be allocated. The memory is set to zero. */

void *__calloc(unsigned int nmemb, unsigned int size) ICODE_ATTR;
void *__calloc(unsigned int nmemb, unsigned int size) {
    void* ret = malloc(size * nmemb);
    if (!ret) return NULL;
    RB memset(ret, 0, size * nmemb);
    return ret;
}

/*  Returns a pointer to a new string which is a duplicate of the
    string _str_. */

char* __strdup(const char* str) ICODE_ATTR;
char* __strdup(const char* str) {
    int l = RB strlen(str) + 1;
    char* ret = __malloc(l);
    if (!ret) 
        return NULL;
    RB memcpy(ret, str, l);
    return ret;
}

void malloc_init(unsigned char *buffer, unsigned int size) {
    mallochead *mallocend;

#ifdef MALLOC_TEST
    unsigned char* alignstart = buffer;
#else
    /* Align to a 16-byte boundary. */
    unsigned char* alignstart = (unsigned char*)(((long)buffer+15)&~15);
#endif
    size -= (alignstart - buffer);

    /* Initialize the free list */
    freelist.next = &freelist;
    freelist.prev = &freelist;
    freelist.psize = 0;
    freelist.size = 0x7fffffff;

    mallocstart = (mallochead *) alignstart;
    heap_size = size;
    size /= 8;

    mallocstart->psize = 0;
    mallocstart->size = size - 1;

    mallocend = ADD(mallocstart, size - 1);

    mallocend->psize = size - 1;
    mallocend->size = 0;

    add_free_chunk(mallocstart);
}

#ifndef MALLOC_TEST

void malloc_heapdump(const char* fname) {
    int fd;

    fd = RB open(fname, O_WRONLY | O_CREAT | O_TRUNC);
    if (fd < 0) {
        return;
    }

    RB write(fd, &mallocstart, sizeof(void*));
    RB write(fd, mallocstart, heap_size);

    RB close(fd);

}

#else

#include <stdio.h>
#include <string.h>

unsigned int mallocbuf[512];

void print_mem(const char* message) {
    int i;
    unsigned int* p = mallocbuf;
    int mode = 0;
    int datacnt = 0;

    if (message)
        printf("%s:\n",message);

    mallochead* m = freelist.next;

    printf("  Free head (%p): %p <-> %p\n", &freelist, freelist.prev, freelist.next);
    while (m != &freelist) {
        printf("   r:%10p - %10p (%4d) free:%10p <-> %10p\n", 
               m, NEXT(m), SIZE(m) - 1 <<3, m->prev, m->next);
        m = m->next;
    }
    printf("  Allocated:\n");
    m = (mallochead*) mallocbuf;
    while (m->size) {
        if (USED(m))
            printf(" X r:%10p - %10p (%4d) lnk: %3d <-> %3d v:\"%s\"\n", 
                   ADD(m, 1), NEXT(m), SIZE(m) - 1 <<3, m->psize, SIZE(m), ADD(m, 1));
        else
            printf("   r:%10p - %10p (%4d) lnk: %3d <-> %3d\n", 
                   ADD(m, 1), NEXT(m), SIZE(m) - 1 <<3, m->psize, SIZE(m));
        m = NEXT(m);
    }
}

void* stralloc(int size, const char* nm) {
    void* ret = __malloc(size);
    if (ret)
        strcpy((char*)ret, nm);
    printf("stralloc(%d,\"%s\") -> %p\n", size, nm, ret); 
    return ret;
}

static inline int randrange(int min, int max) {
    return min + (((unsigned)rand())%(max - min));
}

#define N_PTRS 64

void bang_on_malloc(void) {
    void* ptrs[N_PTRS];
    void* nptr;
    int pm, i, j, k;
    memset(ptrs, 0, sizeof(ptrs));
    k = 0;
    while (1) {
        int ptr = randrange(0, N_PTRS);
        int size = randrange(8, sizeof(mallocbuf)/2);
        pm = 0;
        if (ptrs[ptr]) {
            int op = randrange(0, 2);
            if (op) {
                __free(ptrs[ptr]);
                if (!k)
                    printf("free(ptrs[%d] == %p)\n", ptr, ptrs[ptr]);
                ptrs[ptr] = NULL;
                pm = 1;
            } else {
                nptr = __realloc(ptrs[ptr], size);
                if (!k)
                    printf("realloc(ptrs[%d] == %p, %d) -> %p\n", ptr, ptrs[ptr], size, nptr);
                if (nptr) {
                    ptrs[ptr] = nptr;
                }
                pm = 1;
            }
        } else {
            ptrs[ptr] = __malloc(size);
            if (!k)
                printf("ptrs[%d] = malloc(%d) -> %p\n", ptr, size, ptrs[ptr]);
            if (ptrs[ptr]) {
                snprintf((char*)ptrs[ptr], size, "ptr %d",ptr);
                pm = 1;
            }
        }
        if (pm) {
            if (k) {
                --k;
                continue;
            }
            k = 131072;
            print_mem(NULL);
            for (i = 0; i < N_PTRS/6; i++) {
                for (j = 0; j < 6; j++) {
                    int q = (N_PTRS/6) * j + i;
                    if (q < N_PTRS)
                        printf("%2d = %08x ", q, ptrs[q]);
                }
                printf("\n");
            }
            printf("\n\n");
        }
    }
}


int main(int argc, char** argv) {
    void *foo1, *foo2, *foo3, *foo4, *tmp;
    int i;
    
    srand(42);

    for (i = 0; i < sizeof(mallocbuf)/4; i++) {
	mallocbuf[i] = 0;
    }
    malloc_init((unsigned char*)mallocbuf, sizeof(mallocbuf));

    print_mem("After init");

    bang_on_malloc();

}

#endif
