#!/usr/bin/env python
############################################################################
#             __________               __   ___.
#   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
#   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
#   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
#   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
#                     \/            \/     \/    \/            \/
#
# makerbd.py: Generate RocKen board from GIMP image
#
# Copyright (C) 2006 Katie Stafford
#
# All files in this archive are subject to the GNU General Public License.
# See the file COPYING in the source tree root for full license agreement.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
############################################################################


import math, sys, cStringIO, array, struct, gzip, traceback
from gimpfu import *
from os.path import basename,splitext

def setcell(b, l, x, y, v):
    if x < 0 or y < 0 or x > 63 or y > 63: return
    if v is None: v = 0
    b[l << 12 | x << 6 | y] = v

def setcell_z(b, l, x, y, v):
    if x < 0 or y < 0 or x > 63 or y > 63: return
    if v is None: v = 0
    idx = l << 12 | x << 6 | y
    if b[idx] == 0: b[idx] = v

def dolayer(l, func, bd, o):
    rg = l.get_pixel_rgn(0,0,64,64)
    for x in xrange(64):
        gimp.progress_update(x/64.0)
        for y in xrange(64):
            r,g,b,a = rg[x,y]
            if a == '\x00': continue
            v = "%02x%02x%02x"%(ord(r),ord(g),ord(b))
            func(bd, o, x, y, v)
                
def cellset(map,l,or_=0,xo=0,yo=0):
    def ret(b, o, x, y, v):
        v = map.get(v,None)
        if v is not None:
            setcell_z(b, l, x+xo, y+yo, v|or_)
    return ret

tmap = {
    "808080": 1023,
    "ff0000": 2,
    "0000ff": 1,
    "00ff00": 3,
    "00ffff": 4,
    "ffff00": 5,
    "c53ad6": 6,
    "a9ef23": 7,
    "f9b05a": 8,
    "d84b39": 9,
    "9656ff": 12,
    "b5036c": 10,
    "c9b0f0": 11,
    }

east = cellset(tmap,0,xo=1)
west = cellset(tmap,2,xo=-1)
north = cellset(tmap,3,yo=-1)
south = cellset(tmap,1,yo=1)
cell = cellset(tmap,4,or_=0xc000)

def auto(b, o, x, y, v):
    v = tmap.get(v,None)

    if v is None: return
    
    setcell_z(b, 4, x, y, v|0xc000);
    setcell_z(b, 0, x+1, y, v);
    setcell_z(b, 1, x, y+1, v);
    setcell_z(b, 2, x-1, y, v);
    setcell_z(b, 3, x, y-1, v);

def obj(**kw):
    def ret(b, o, x, y, v):
        nv = kw.get('x'+v,None)
        if nv is None:
            #sys.stderr.write(repr((x,y,v,kw))+'\n')
            return
        nv(b, o, x, y, v)
    return ret

def writestr(o, str):
    o.write(struct.pack(">H",len(str)))
    o.write(str);

# Define new objects here.

def writeactor(o, clas, x, y, ang, fmt = "", *args):
    if x < 64:
        x = x<<16 | 32768
    if y < 64:
        y = y<<16 | 32768
    writestr(o, clas)
    o.write(struct.pack(">IIH"+fmt, x, y, ang, *args));
    
def door(dir, nk=0):
    def ret(b, o, x, y, v):
        writeactor(o, "door", x, y, 0, "HIBB", 0, 0, dir, nk),
    return ret

def secretwall(dir):
    def ret(b, o, x, y, v):
        writeactor(o, "secretwall", x, y, 0, "B", dir);
    return ret

def player(ang):
    def ret(b, o, x, y, v):
        writeactor(o, "player", x, y, ang);
    return ret

def robot(b, o, x, y, v):
    writeactor(o, "robot", x, y, 0);

def alien(b, o, x, y, v):
    writeactor(o, "alien", x, y, 0);

def key(b, o, x, y, v):
    writeactor(o, "key", x, y, 0);

def medkit(amt):
    def ret(b, o, x, y, v):
        writeactor(o, "medkit", x, y, 0, "B", amt);
    return ret

def map(b, o, x, y, v):
    writeactor(o, "map", x, y, 0);

def generator(clas,min_intv,max_intv,nobj):
    def ret(b, o, x, y, v):
        od = cStringIO.StringIO()
        clas(b, od, x, y, v)
        v = od.getvalue()
        writeactor(o, "generator", x, y, 0, "HHH", min_intv, max_intv, nobj);
        writestr(o, od.getvalue());
    return ret


def make_rbd(timg, tdrawable):
    fname = timg.filename
    ofname = splitext(fname)[0] + '.rbd'

    board=[0]*(4096*5);
    objdata = cStringIO.StringIO()



    for l in timg.layers:
        try:
            gimp.progress_init("processing layer %s"%l.name);
            foo=eval(l.name+'\n')
        except:
            print >>sys.stderr, "WARNING:"
            traceback.print_exc(sys.stderr)
            continue
        if callable(foo):
            dolayer(l, foo, board, objdata)
    gimp.progress_init("writing %s"%ofname);
    fil=gzip.GzipFile(ofname,'wb')
    fil.write(struct.pack(">20480H",*board))
    fil.write(objdata.getvalue())
    gimp.progress_init("done!");
        

register(
        "python_fu_make_rbd",
        "Export image to RocKen board",
        "Export image to RocKen board",
        "Katie Stafford",
        "Katie Stafford",
        "2006-2006",
        "<Image>/File/Export RocKen board",
        "RGB*",
        [
        ],
        [],
        make_rbd)

main()
